########################################################################################################################
# Module: ssm/scenarios/lorenz96.py
# Description: Latent Lorenz96 dynamics with TemporalGaussian transition and observation noise.
#
# Web: https://github.com/SamDuffield/mocat
########################################################################################################################

from jax import numpy as jnp, jit
from jax.experimental.ode import odeint

from mocat.src.ssm.nonlinear_gaussian import NonLinearGaussian


def lorenz96_dynamics(x: jnp.ndarray,
                      t: float,
                      forcing_constant: float) -> jnp.ndarray:
    d = len(x)
    return (x[(jnp.arange(d) + 1) % d] - x[(jnp.arange(d) - 2) % d]) * x[(jnp.arange(d) - 1) % d] \
           - x + forcing_constant


@jit
def lorenz96_integrator(x: jnp.ndarray,
                        delta_t: float,
                        forcing_constant: float) -> jnp.ndarray:
    return odeint(lorenz96_dynamics, x, jnp.array([0, delta_t]), forcing_constant)[-1]


class Lorenz96(NonLinearGaussian):
    name = 'Lorenz 96'

    def __init__(self,
                 dim: int = 40,
                 forcing_constant: float = 8.,
                 **kwargs):
        self.dim = dim
        self.forcing_constant = forcing_constant
        super().__init__(**kwargs)

    def transition_function(self,
                            x_previous: jnp.ndarray,
                            t_previous: float,
                            t_new: float) -> jnp.ndarray:
        return lorenz96_integrator(x_previous, t_new - t_previous, self.forcing_constant)



########################################################################################################################
# Module: ssm/tests/test_ssm_linear_gaussian.py
# Description: Tests for 1D non-linear benchmark model.
#
# Web: https://github.com/SamDuffield/mocat
########################################################################################################################


import unittest

from mocat.src.tests.test_ssm import TestSSM
from mocat.src.ssm.scenarios.nonlinear1d import NonLinear1DBenchmark


class Test1DBootstrap(TestSSM):

    ssm_scenario = NonLinear1DBenchmark()

    def test_simulate(self):
        super()._test_simulate()

    def test_bootstrap(self):
        super()._test_bootstrap()

    def test_backward(self):
        super()._test_backward()

    def test_online_smoothing_pf(self):
        super()._test_online_smoothing_pf_full()
        super()._test_online_smoothing_pf_rejection()

    def test_online_smoothing_bs(self):
        super()._test_online_smoothing_bs_full()
        super()._test_online_smoothing_bs_rejection()


if __name__ == '__main__':
    unittest.main()



########################################################################################################################
# Module: transport/teki.py
# Description: Tempered ensemble Kalman inversion for intractable likelihoods.
#
# Web: https://github.com/SamDuffield/mocat
########################################################################################################################


from typing import Tuple, Union, Callable

from jax import numpy as jnp, random, vmap

from mocat.src.core import Scenario, cdict
from mocat.src.transport.sampler import TransportSampler
from mocat.src.utils import gaussian_potential, bisect
from mocat.src.metrics import log_ess_log_weight


def calculate_covariances(vals: jnp.ndarray,
                          simulated_data: jnp.ndarray,
                          mean_vals: jnp.ndarray = None,
                          mean_sim_data: jnp.ndarray = None) -> Tuple[jnp.ndarray, jnp.ndarray, jnp.ndarray]:
    n = vals.shape[0]
    if mean_vals is None:
        mean_vals = vals.mean(0)
    if mean_sim_data is None:
        mean_sim_data = simulated_data.mean(0)

    s_x = (vals - mean_vals) / jnp.sqrt(n - 1)
    s_y = (simulated_data - mean_sim_data) / jnp.sqrt(n - 1)

    cov_x = s_x.T @ s_x
    cov_xy = s_x.T @ s_y
    cov_y = s_y.T @ s_y

    return cov_x, cov_xy, cov_y


class TemperedEKI(TransportSampler):
    name = "Tempered EKI"

    def __init__(self,
                 temperature_schedule: Union[None, float, jnp.ndarray] = None,
                 next_temperature: Callable = None,
                 max_temperature: float = 1.,
                 max_iter: int = int(1e4),
                 term_std: float = 0.,
                 nugget: float = 1e-5,
                 **kwargs):
        self.max_iter = max_iter
        self.max_temperature = max_temperature
        self.temperature_schedule = temperature_schedule
        super().__init__(**kwargs)
        self.parameters.nugget = nugget
        self.parameters.term_std = term_std
        if next_temperature is not None:
            self.next_temperature = next_temperature
        self.ensemble_std = None
        self.prior_stds = None

    def __setattr__(self, key, value):
        super().__setattr__(key, value)
        if key == 'temperature_schedule':
            if value is not None:
                self.max_temperature = value[-1]
                self.max_iter = len(value)
                self.next_temperature = lambda state, extra: self.temperature_schedule[extra.iter]

    def startup(self,
                scenario: Scenario,
                n: int,
                initial_state: cdict,
                initial_extra: cdict,
                **kwargs) -> Tuple[cdict, cdict]:
        initial_state, initial_extra = super().startup(scenario, n, initial_state, initial_extra, **kwargs)

        if not hasattr(scenario, 'data') or scenario.data is None:
            raise AttributeError(f'{self.name} requires scenario to have data attribute != None')

        if self.ensemble_std is None:
            ensemble_std = lambda vals: jnp.std(vals, axis=0, ddof=1)
            if hasattr(scenario, 'constrain'):
                self.ensemble_std = lambda vals: ensemble_std(scenario.constrain(vals))
            else:
                self.ensemble_std = ensemble_std

        if not hasattr(self, 'next_temperature'):
            # gamm = scenario.dim ** -0.5
            # self.next_temperature = lambda state, extra: (1 + gamm) ** extra.iter - 1

            gamm = 2 ** (1 / 50)
            self.next_temperature = lambda state, extra: jnp.round(gamm ** extra.iter - 1, 4)

        if not hasattr(initial_state, 'simulated_data'):
            random_keys = random.split(initial_extra.random_key, 2 * n)
            initial_extra.random_key = random_keys[-1]
            initial_state.simulated_data = vmap(scenario.likelihood_sample)(initial_state.value,
                                                                            random_keys[:n])

        self.prior_stds = jnp.std(initial_state.value, axis=0, ddof=1)

        initial_state.temperature = 0.
        initial_extra.data = scenario.data
        initial_extra.prec_y_given_x = jnp.eye(initial_state.simulated_data.shape[-1])
        initial_state.perturb_nan = 0
        return initial_state, initial_extra

    def termination_criterion(self,
                              ensemble_state: cdict,
                              extra: cdict) -> bool:
        return jnp.any(jnp.array([ensemble_state.temperature >= self.max_temperature,
                                  extra.iter >= self.max_iter,
                                  jnp.all(self.ensemble_std(ensemble_state.value) <
                                          (extra.parameters.term_std * self.prior_stds)),
                                  jnp.any(jnp.isnan(ensemble_state.value))]))

    def update(self,
               scenario: Scenario,
               ensemble_state: cdict,
               extra: cdict) -> Tuple[cdict, cdict]:
        extra.iter += 1
        n, d_x = ensemble_state.value.shape
        d_y = ensemble_state.simulated_data.shape[-1]
        random_keys = random.split(extra.random_key, n + 2)
        extra.random_key = random_keys[-1]

        cov_x, cov_xy, cov_y = calculate_covariances(ensemble_state.value, ensemble_state.simulated_data)
        cov_y_given_x = cov_y - cov_xy.T @ jnp.linalg.inv(cov_x + self.parameters.nugget * jnp.eye(d_x)) @ cov_xy
        cov_y_given_x_chol = jnp.linalg.cholesky(cov_y_given_x + self.parameters.nugget * jnp.eye(d_y))
        extra.prec_y_given_x = jnp.linalg.inv(cov_y_given_x + self.parameters.nugget * jnp.eye(d_y))

        prev_temp = ensemble_state.temperature
        new_temp = self.next_temperature(ensemble_state, extra)
        alph = 1. / (new_temp - prev_temp)
        ensemble_state.temperature = new_temp

        cov_alph = cov_y + (alph - 1) * cov_y_given_x
        kalman_gain = cov_xy @ jnp.linalg.inv(cov_alph + self.parameters.nugget * jnp.eye(d_y))

        perturbs = jnp.sqrt(alph - 1) * random.normal(random_keys[-2], shape=(n, d_y)) @ cov_y_given_x_chol.T
        ensemble_state.perturb_nan = jnp.isnan(perturbs).sum()
        perturbs = jnp.where(jnp.isnan(perturbs), 0., perturbs)

        ensemble_state.value = ensemble_state.value \
                               + (scenario.data - ensemble_state.simulated_data + perturbs) @ kalman_gain.T

        ensemble_state.simulated_data = vmap(scenario.likelihood_sample)(ensemble_state.value,
                                                                         random_keys[:n])

        return ensemble_state, extra


class AdaptiveTemperedEKI(TemperedEKI):

    def __init__(self,
                 max_temperature: float = 1.,
                 max_iter: int = int(1e4),
                 nugget: float = 1e-5,
                 term_std: float = 0.,
                 ess_threshold: float = 0.9,
                 bisection_tol: float = 1e-5,
                 max_bisection_iter: int = 1000,
                 **kwargs):
        super().__init__(temperature_schedule=None,
                         max_temperature=max_temperature, max_iter=max_iter,
                         nugget=nugget, term_std=term_std, **kwargs)
        self.parameters.ess_threshold = ess_threshold
        self.parameters.bisection_tol = bisection_tol
        self.parameters.max_bisection_iter = max_bisection_iter

    def next_temperature(self,
                         ensemble_state: cdict,
                         extra: cdict) -> float:
        temperature_bounds = jnp.array([ensemble_state.temperature, self.max_temperature])
        data_diff = ensemble_state.simulated_data - extra.data
        pseudo_likelihood_potential = vmap(lambda diffi: 0.5 * diffi.T @ extra.prec_y_given_x @ diffi)(data_diff)
        log_n_samp_threshold = jnp.log(len(pseudo_likelihood_potential) * self.parameters.ess_threshold)
        bisect_out_bounds, bisect_out_evals, bisect_out_iter \
            = bisect(lambda x: log_ess_log_weight(- (x - ensemble_state.temperature) * pseudo_likelihood_potential)
                               - log_n_samp_threshold,
                     temperature_bounds,
                     max_iter=self.parameters.max_bisection_iter,
                     tol=self.parameters.bisection_tol)

        return bisect_out_bounds[jnp.argmin(jnp.abs(bisect_out_evals))]



########################################################################################################################
# Module: tests/test_kernels.py
# Description: Tests for kernels
#
# Web: https://github.com/SamDuffield/mocat
########################################################################################################################

import unittest

import jax.numpy as jnp
import numpy.testing as npt

from mocat.src import kernels


class TestGaussianKernel(unittest.TestCase):
    kernel = kernels.Gaussian()

    def test_call(self):
        npt.assert_array_almost_equal(self.kernel(jnp.zeros(5), jnp.zeros(5)), 1.)
        npt.assert_array_almost_equal(self.kernel(jnp.zeros(5), jnp.ones(5)), 0.082085006)

    def test_grad_x(self):
        npt.assert_array_almost_equal(self.kernel.grad_x(jnp.zeros(5), jnp.zeros(5)), jnp.zeros(5))
        npt.assert_array_almost_equal(self.kernel.grad_x(jnp.zeros(5), jnp.ones(5)), jnp.ones(5) * 0.082085006)

    def test_grad_y(self):
        npt.assert_array_almost_equal(self.kernel.grad_y(jnp.zeros(5), jnp.zeros(5)), jnp.zeros(5))
        npt.assert_array_almost_equal(self.kernel.grad_y(jnp.zeros(5), jnp.ones(5)), jnp.ones(5) * -0.082085006)


class TestIMQKernel(unittest.TestCase):
    kernel = kernels.IMQ()

    def test_call(self):
        npt.assert_array_almost_equal(self.kernel(jnp.zeros(5), jnp.zeros(5)), 1.)
        npt.assert_array_almost_equal(self.kernel(jnp.zeros(5), jnp.ones(5)), 0.5345225)

    def test_grad_x(self):
        npt.assert_array_almost_equal(self.kernel.grad_x(jnp.zeros(5), jnp.zeros(5)), jnp.zeros(5))
        npt.assert_array_almost_equal(self.kernel.grad_x(jnp.zeros(5), jnp.ones(5)), jnp.ones(5) * 0.07636035)

    def test_grad_y(self):
        npt.assert_array_almost_equal(self.kernel.grad_y(jnp.zeros(5), jnp.zeros(5)), jnp.zeros(5))
        npt.assert_array_almost_equal(self.kernel.grad_y(jnp.zeros(5), jnp.ones(5)), jnp.ones(5) * -0.07636035)


if __name__ == '__main__':
    unittest.main()




from mocat.src.transport.sampler import TransportSampler

from mocat.src.transport.smc import SMCSampler
from mocat.src.transport.smc import TemperedSMCSampler
from mocat.src.transport.smc import MetropolisedSMCSampler
from mocat.src.transport.smc import RMMetropolisedSMCSampler

from mocat.src.transport.svgd import SVGD
from mocat.src.transport.svgd import kernelised_grad_matrix





########################################################################################################################
# Module: tests/test_metrics.py
# Description: Tests for metrics analysing run quality.
#
# Web: https://github.com/SamDuffield/mocat
########################################################################################################################

import unittest

import jax.numpy as jnp
from jax import random
import numpy.testing as npt

from mocat.src.core import cdict
from mocat.src import kernels, metrics


class TestAutocorrelation(unittest.TestCase):
    key = random.PRNGKey(0)
    ind_draws_arr = random.normal(key, (100,))
    ind_draws_cdict = cdict(value=ind_draws_arr)

    corr_draws_arr = 0.1 * jnp.cumsum(ind_draws_arr)
    corr_draws_cdict = cdict(value=corr_draws_arr)

    ind_draws_cdict_pot = cdict(value=corr_draws_arr,
                                potential=ind_draws_arr)
    corr_draws_cdict_pot = cdict(value=ind_draws_arr,
                                 potential=corr_draws_arr)

    def test_array(self):
        ind_autocorr = metrics.autocorrelation(self.ind_draws_arr)
        self.assertEqual(ind_autocorr[0], 1.)
        npt.assert_array_equal(jnp.abs(ind_autocorr[1:]) < 0.3, True)

        corr_autocorr = metrics.autocorrelation(self.corr_draws_arr)
        self.assertEqual(corr_autocorr[0], 1.)
        npt.assert_array_equal(jnp.abs(corr_autocorr[1]) > 0.5, True)
        npt.assert_array_equal(jnp.abs(corr_autocorr[50]) < 0.5, True)
        npt.assert_array_equal(jnp.abs(corr_autocorr[-1]) < 0.3, True)

    def test_cdict_pot(self):
        # Potential
        ind_autocorr = metrics.autocorrelation(self.ind_draws_cdict_pot)
        self.assertEqual(ind_autocorr[0], 1.)
        npt.assert_array_equal(jnp.abs(ind_autocorr[1:]) < 0.3, True)

        corr_autocorr = metrics.autocorrelation(self.corr_draws_cdict_pot)
        self.assertEqual(corr_autocorr[0], 1.)
        npt.assert_array_equal(jnp.abs(corr_autocorr[1]) > 0.5, True)
        npt.assert_array_equal(jnp.abs(corr_autocorr[50]) < 0.5, True)
        npt.assert_array_equal(jnp.abs(corr_autocorr[-1]) < 0.3, True)


class testIAT(unittest.TestCase):
    key = random.PRNGKey(0)
    ind_draws_arr = random.normal(key, (100,))
    ind_draws_cdict = cdict(value=ind_draws_arr)

    corr_draws_arr = 0.1 * jnp.cumsum(ind_draws_arr)

    def test_array(self):
        ind_iat = metrics.integrated_autocorrelation_time(self.ind_draws_arr)
        corr_iat = metrics.integrated_autocorrelation_time(self.corr_draws_arr)
        self.assertLess(ind_iat, 2.)
        self.assertGreater(corr_iat, 8.)


class testSJD(unittest.TestCase):
    zeros_arr = jnp.zeros(10)
    zeros_cdict = cdict(value=zeros_arr)
    seq_arr = jnp.arange(10)
    seq_cdict = cdict(value=seq_arr)

    def test_array(self):
        accept_rate = metrics.squared_jumping_distance(self.zeros_arr)
        self.assertEqual(accept_rate, 0.)

        accept_rate = metrics.squared_jumping_distance(self.seq_arr)
        self.assertEqual(accept_rate, 1.)

    def test_cdict(self):
        accept_rate = metrics.squared_jumping_distance(self.zeros_cdict)
        self.assertEqual(accept_rate, 0.)

        accept_rate = metrics.squared_jumping_distance(self.seq_cdict)
        self.assertEqual(accept_rate, 1.)


class testKSDStdGaussian(unittest.TestCase):
    key = random.PRNGKey(0)
    dim = 2

    n_small = 10
    ind_draws_arr_n_small = random.normal(key, (n_small, dim))
    sample_n_small = cdict(value=ind_draws_arr_n_small, grad_potential=ind_draws_arr_n_small)

    n_large = 1000
    ind_draws_arr_n_large = random.normal(key, (n_large, dim))
    sample_n_large = cdict(value=ind_draws_arr_n_large, grad_potential=ind_draws_arr_n_large)

    def testksd_gaussian_kernel(self):
        kernel = kernels.Gaussian(bandwidth=1.)
        ksd_n_small_a = metrics.ksd(self.sample_n_small, kernel)
        ksd_n_large_a = metrics.ksd(self.sample_n_large, kernel)
        ksd_n_large_a_minibatch = metrics.ksd(self.sample_n_small, kernel, ensemble_batchsize=100, random_key=self.key)

        self.assertLess(ksd_n_large_a, ksd_n_small_a)
        npt.assert_almost_equal(ksd_n_large_a, ksd_n_large_a_minibatch, 1)

        kernel.parameters.bandwidth = 10.
        ksd_n_small_b = metrics.ksd(self.sample_n_small, kernel)
        ksd_n_large_b = metrics.ksd(self.sample_n_large, kernel)

        self.assertLess(ksd_n_large_b, ksd_n_small_b)

    def testksd_IMQ_kernel(self):
        kernel = kernels.IMQ(bandwidth=1., c=1., beta=-0.5)
        ksd_n_small_a = metrics.ksd(self.sample_n_small, kernel)
        ksd_n_large_a = metrics.ksd(self.sample_n_large, kernel)
        ksd_n_large_a_minibatch = metrics.ksd(self.sample_n_small, kernel, ensemble_batchsize=100, random_key=self.key)

        self.assertLess(ksd_n_large_a, ksd_n_small_a)
        npt.assert_almost_equal(ksd_n_large_a, ksd_n_large_a_minibatch, 1)

        kernel.parameters.bandwidth = 10.
        kernel.parameters.c = 0.1
        kernel.parameters.beta = -0.1

        ksd_n_small_b = metrics.ksd(self.sample_n_small, kernel)
        ksd_n_large_b = metrics.ksd(self.sample_n_large, kernel)

        self.assertLess(ksd_n_large_b, ksd_n_small_b)


if __name__ == '__main__':
    unittest.main()




from mocat import twodim

from mocat.src.core import Scenario

from mocat.src.scenarios.toy_examples import Gaussian
from mocat.src.scenarios.toy_examples import GaussianMixture
from mocat.src.scenarios.toy_examples import DoubleWell
from mocat.src.scenarios.toy_examples import Rastrigin







from setuptools import setup, find_namespace_packages

exec(open('mocat/version.py').read())

with open('README.md') as f:
    long_description = f.read()

setup(
    name='mocat',
    version=__version__,
    url='http://github.com/SamDuffield/mocat',
    author='Sam Duffield',
    python_requires='>=3.6',
    install_requires=['jax',
                      'jaxlib',
                      'matplotlib',
                      'decorator',
                      'numpy'],
    packages=find_namespace_packages(),
    author_email='sddd2@cam.ac.uk',
    description='Monte Carlo Testbed',
    long_description=long_description,
    long_description_content_type='text/markdown',
    license='MIT'
)





########################################################################################################################
# Module: scenarios/twodim/toy_examples.py
# Description: Initiates some toy 2D scenarios suitable for visualisation.
#
# Web: https://github.com/SamDuffield/mocat
########################################################################################################################

from typing import Union

import jax.numpy as jnp

from mocat.src.scenarios.twodim.vectorise import TwoDimToyScenario
from mocat.src.scenarios.toy_examples import Gaussian as NDGaussian
from mocat.src.scenarios.toy_examples import GaussianMixture as NDGaussianMixture
from mocat.src.scenarios.toy_examples import DoubleWell as NDDoubleWell
from mocat.src.scenarios.toy_examples import Rastrigin as NDRastrigin


class Gaussian(TwoDimToyScenario, NDGaussian):
    def __init__(self,
                 mean: jnp.ndarray = jnp.array([0., 0.]),
                 covariance: jnp.ndarray = None):
        TwoDimToyScenario.__init__(self)
        NDGaussian.__init__(self, 2, mean, covariance)


class Banana(TwoDimToyScenario):
    name = 'Banana'
    curviness = 0.03
    lengthiness = 100

    def likelihood_potential(self,
                             x: jnp.ndarray,
                             random_key: jnp.ndarray = None) -> Union[float, jnp.ndarray]:
        return 0.5 * (x[0] ** 2 / self.lengthiness + (x[1] + self.curviness * x[0] ** 2
                                                      - self.lengthiness * self.curviness) ** 2)


class NealFunnel(TwoDimToyScenario):
    name = "Neal's Funnel"

    def likelihood_potential(self,
                             x: jnp.ndarray,
                             random_key: jnp.ndarray = None) -> Union[float, jnp.ndarray]:
        return 0.5 * (x[1] ** 2 / 3 + x[0] ** 2 * jnp.exp(-x[1] / 2))


class GaussianMixture(TwoDimToyScenario, NDGaussianMixture):
    def __init__(self,
                 means: jnp.ndarray = jnp.array([[-2, -2], [2, -2], [-2, 2], [2, 2]]),
                 **kwargs):
        TwoDimToyScenario.__init__(self)
        NDGaussianMixture.__init__(self, means, **kwargs)


class DoubleWell(TwoDimToyScenario, NDDoubleWell):
    def __init__(self):
        TwoDimToyScenario.__init__(self)
        NDDoubleWell.__init__(self, dim=2)


class Rastrigin(TwoDimToyScenario, NDRastrigin):
    def __init__(self,
                 a: float = 1.):
        TwoDimToyScenario.__init__(self)
        NDRastrigin.__init__(self, dim=2, a=a)





########################################################################################################################
# Module: mcmc/sampler.py
# Description: Generic MCMC sampler class.
#
# Web: https://github.com/SamDuffield/mocat
########################################################################################################################

from typing import Tuple, Union, Type
from inspect import isclass

from jax import numpy as jnp, random

from mocat.src.core import Scenario, cdict, is_implemented
from mocat.src.sample import Sampler


class MCMCSampler(Sampler):
    correction: 'Correction' = None

    def __init__(self,
                 **kwargs):

        if not hasattr(self, 'tuning'):
            self.tuning = cdict(parameter='stepsize',
                                target=None,
                                metric='alpha',
                                monotonicity='decreasing')

        # Initiate sampler class (set any additional parameters from init)
        super().__init__(**kwargs)

    def startup(self,
                scenario: Scenario,
                n: int,
                initial_state: cdict,
                initial_extra: cdict,
                startup_correction: bool = True,
                **kwargs) -> Tuple[cdict, cdict]:
        if initial_state is None:
            if is_implemented(scenario.prior_sample):
                initial_extra.random_key, sub_key = random.split(initial_extra.random_key)
                init_vals = scenario.prior_sample(sub_key)
            else:
                init_vals = jnp.zeros(scenario.dim)
            initial_state = cdict(value=init_vals)

        self.max_iter = n - 1

        if 'correction' in kwargs.keys():
            self.correction = kwargs['correction']
            del kwargs['correction']

        self.correction = check_correction(self.correction)

        initial_state, initial_extra = super().startup(scenario, n, initial_state, initial_extra, **kwargs)

        if startup_correction:
            initial_state, initial_extra = self.correction.startup(scenario, self, n,
                                                                   initial_state, initial_extra, **kwargs)

        return initial_state, initial_extra

    def clean_chain(self,
                    scenario: Scenario,
                    chain_state: cdict) -> cdict:
        return self.correction.clean_chain(scenario, self, chain_state)

    def always(self,
               scenario: Scenario,
               reject_state: cdict,
               reject_extra: cdict) -> Tuple[cdict, cdict]:
        return reject_state, reject_extra

    def proposal(self,
                 scenario: Scenario,
                 reject_state: cdict,
                 reject_extra: cdict) -> Tuple[cdict, cdict]:
        raise NotImplementedError(f'{self.__class__.__name__} markov proposal not initiated')

    def forward_potential(self,
                           scenario: Scenario,
                           reject_state: cdict, reject_extra: cdict,
                           proposed_state: cdict, proposed_extra: cdict) -> Union[float, jnp.ndarray]:
        raise AttributeError(f'{self.__class__.__name__} forward_potential not initiated')

    def acceptance_probability(self,
                               scenario: Scenario,
                               reject_state: cdict, reject_extra: cdict,
                               proposed_state: cdict, proposed_extra: cdict) -> Union[float, jnp.ndarray]:
        raise AttributeError(f'{self.__class__.__name__} acceptance_probability not initiated')

    def update(self,
               scenario: Scenario,
               previous_state: cdict,
               previous_extra: cdict) -> Tuple[cdict, cdict]:

        previous_extra.iter += 1

        reject_state, reject_extra = self.always(scenario,
                                                 previous_state,
                                                 previous_extra)

        proposed_state, proposed_extra = self.proposal(scenario,
                                                       reject_state,
                                                       reject_extra)

        corrected_state, corrected_extra = self.correction(scenario, self,
                                                           reject_state, reject_extra,
                                                           proposed_state, proposed_extra)

        return corrected_state, corrected_extra


class Correction:

    def __init__(self, **kwargs):
        self.__dict__.update(kwargs)

    def __repr__(self):
        return f'mocat.Correction.{self.__class__.__name__}'

    def startup(self,
                scenario: Scenario,
                sampler: MCMCSampler,
                n: int,
                initial_state: cdict,
                initial_extra: cdict,
                **kwargs) -> Tuple[cdict, cdict]:
        # Update kwargs
        for key, value in kwargs.items():
            if hasattr(self, key):
                setattr(self, key, value)
        return initial_state, initial_extra

    def correct(self,
                scenario: Scenario,
                sampler: MCMCSampler,
                reject_state: cdict,
                reject_extra: cdict,
                proposed_state: cdict,
                proposed_extra: cdict) -> Tuple[cdict, cdict]:
        raise NotImplementedError

    def __call__(self,
                 scenario: Scenario,
                 sampler: MCMCSampler,
                 reject_state: cdict,
                 reject_extra: cdict,
                 proposed_state: cdict,
                 proposed_extra: cdict) -> Tuple[cdict, cdict]:
        return self.correct(scenario, sampler,
                            reject_state, reject_extra,
                            proposed_state, proposed_extra)

    def clean_chain(self,
                    scenario: Scenario,
                    sampler: Sampler,
                    chain_state: cdict) -> cdict:
        return chain_state


class Uncorrected(Correction):

    def correct(self,
                scenario: Scenario,
                sampler: MCMCSampler,
                reject_state: cdict,
                reject_extra: cdict,
                proposed_state: cdict,
                proposed_extra: cdict) -> Tuple[cdict, cdict]:
        return proposed_state, proposed_extra


def check_correction(correction: Union[None, Correction, Type[Correction]]) -> Correction:
    if correction is None:
        correction = Uncorrected()
    elif isclass(correction) and issubclass(correction, Correction):
        correction = correction()
    elif not isinstance(correction, Correction):
        raise TypeError(f'Correction must be of type mocat.Correction')

    return correction



########################################################################################################################
# Module: kernels.py
# Description: Kernel class definition plus some standard kernels.
#
# Web: https://github.com/SamDuffield/mocat
########################################################################################################################

from typing import Union
from functools import partial

from jax import jit, numpy as jnp

from mocat.src.core import cdict
from mocat.src.utils import l2_distance_matrix


class Kernel:
    parameters: cdict

    def __init__(self, **kwargs):
        if not hasattr(self, 'parameters'):
            self.parameters = cdict()
        self.parameters.__dict__.update(kwargs)

    def __repr__(self):
        return f"mocat.kernels.{self.__class__.__name__}"

    def _call(self,
              x: jnp.ndarray,
              y: jnp.ndarray,
              **kernel_params) -> Union[float, jnp.ndarray]:
        raise NotImplementedError(f'kernel.{self.__class__.__name__} call not implemented')

    def _grad_x(self,
                x: jnp.ndarray,
                y: jnp.ndarray,
                **kernel_params) -> Union[float, jnp.ndarray]:
        raise NotImplementedError(f'kernel.{self.__class__.__name__} grad_x not implemented')

    def _grad_y(self,
                x: jnp.ndarray,
                y: jnp.ndarray,
                **kernel_params) -> Union[float, jnp.ndarray]:
        raise NotImplementedError(f'kernel.{self.__class__.__name__} grad_x not implemented')

    def _diag_grad_xy(self,
                      x: jnp.ndarray,
                      y: jnp.ndarray,
                      **kernel_params) -> Union[float, jnp.ndarray]:
        raise NotImplementedError(f'kernel.{self.__class__.__name__} diag_grad_xy not implemented')

    def _flex_params(self, ijnput_params: dict) -> dict:
        kern_params = self.parameters.__dict__.copy()
        kern_params.update(ijnput_params)
        return kern_params

    def __call__(self,
                 x: jnp.ndarray,
                 y: jnp.ndarray,
                 **kwargs) -> Union[float, jnp.ndarray]:
        return self._call(x, y, **self._flex_params(kwargs))

    def grad_x(self,
               x: jnp.ndarray,
               y: jnp.ndarray,
               **kwargs) -> Union[float, jnp.ndarray]:
        return self._grad_x(x, y, **self._flex_params(kwargs))

    def grad_y(self,
               x: jnp.ndarray,
               y: jnp.ndarray,
               **kwargs) -> Union[float, jnp.ndarray]:
        return self._grad_y(x, y, **self._flex_params(kwargs))

    def diag_grad_xy(self,
                     x: jnp.ndarray,
                     y: jnp.ndarray,
                     **kwargs) -> Union[float, jnp.ndarray]:
        return self._diag_grad_xy(x, y, **self._flex_params(kwargs))


class Gaussian(Kernel):

    def __init__(self,
                 bandwidth: Union[float, jnp.ndarray] = 1.):
        super().__init__()
        self.parameters.bandwidth = bandwidth

    @partial(jit, static_argnums=(0,))
    def _call(self,
              x: jnp.ndarray,
              y: jnp.ndarray,
              bandwidth: Union[float, jnp.ndarray]) -> Union[float, jnp.ndarray]:
        diff = (x - y) / bandwidth
        return jnp.exp(-0.5 * jnp.sum(jnp.square(diff), axis=-1))

    @partial(jit, static_argnums=(0,))
    def _grad_x(self,
                x: jnp.ndarray,
                y: jnp.ndarray,
                bandwidth: Union[float, jnp.ndarray]) -> Union[float, jnp.ndarray]:
        return (y - x) * self._call(x, y, bandwidth=bandwidth) / bandwidth ** 2

    @partial(jit, static_argnums=(0,))
    def _grad_y(self,
                x: jnp.ndarray,
                y: jnp.ndarray,
                bandwidth: Union[float, jnp.ndarray]) -> Union[float, jnp.ndarray]:
        return (x - y) * self._call(x, y, bandwidth=bandwidth) / bandwidth ** 2

    @partial(jit, static_argnums=(0,))
    def _diag_grad_xy(self,
                      x: jnp.ndarray,
                      y: jnp.ndarray,
                      bandwidth: Union[float, jnp.ndarray]) -> Union[float, jnp.ndarray]:
        return (bandwidth ** 2 - (x - y) ** 2) * self._call(x, y, bandwidth=bandwidth) / bandwidth ** 4


class PreconditionedGaussian(Kernel):

    def __init__(self,
                 bandwidth: float = 1.,
                 precision: jnp.ndarray = None):
        super().__init__()
        self.parameters.bandwidth = bandwidth
        self.parameters.precision = precision

    @partial(jit, static_argnums=(0,))
    def _call(self,
              x: jnp.ndarray,
              y: jnp.ndarray,
              bandwidth: float,
              precision: jnp.ndarray) -> Union[float, jnp.ndarray]:
        diff = (x - y) / bandwidth
        return jnp.exp(-0.5 * jnp.sum(diff.T @ precision @ diff))

    @partial(jit, static_argnums=(0,))
    def _grad_x(self,
                x: jnp.ndarray,
                y: jnp.ndarray,
                bandwidth: float,
                precision: jnp.ndarray) -> Union[float, jnp.ndarray]:
        return precision @ (y - x) / bandwidth ** 2 * self._call(x, y, bandwidth=bandwidth, precision=precision)

    @partial(jit, static_argnums=(0,))
    def _grad_y(self,
                x: jnp.ndarray,
                y: jnp.ndarray,
                bandwidth: float,
                precision: jnp.ndarray) -> Union[float, jnp.ndarray]:
        return precision @ (x - y) / bandwidth ** 2 * self._call(x, y, bandwidth=bandwidth, precision=precision)

    @partial(jit, static_argnums=(0,))
    def _diag_grad_xy(self,
                      x: jnp.ndarray,
                      y: jnp.ndarray,
                      bandwidth: float,
                      precision: jnp.ndarray) -> Union[float, jnp.ndarray]:
        return precision / bandwidth ** 2 @ (1 - precision * ((x - y) / bandwidth) ** 2) \
               * self._call(x, y,
                            bandwidth=bandwidth,
                            precision=precision)


class IMQ(Kernel):

    def __init__(self,
                 c: Union[float, jnp.ndarray] = 1.,
                 beta: Union[float, jnp.ndarray] = -0.5,
                 bandwidth: Union[float, jnp.ndarray] = 1.):
        super().__init__()
        self.parameters.c = c
        self.parameters.beta = beta
        self.parameters.bandwidth = bandwidth

    @partial(jit, static_argnums=(0,))
    def _call(self,
              x: jnp.ndarray,
              y: jnp.ndarray,
              c: Union[float, jnp.ndarray],
              beta: Union[float, jnp.ndarray],
              bandwidth: Union[float, jnp.ndarray]) -> Union[float, jnp.ndarray]:
        diff = (x - y) / bandwidth
        return jnp.power(c + 0.5 * jnp.sum(jnp.square(diff), axis=-1), beta)

    @partial(jit, static_argnums=(0,))
    def _grad_x(self,
                x: jnp.ndarray,
                y: jnp.ndarray,
                c: Union[float, jnp.ndarray],
                beta: Union[float, jnp.ndarray],
                bandwidth: Union[float, jnp.ndarray]) -> Union[float, jnp.ndarray]:
        diff = (x - y) / bandwidth
        return diff * beta * jnp.power(c + 0.5 * jnp.sum(jnp.square(diff), axis=-1), beta - 1)

    @partial(jit, static_argnums=(0,))
    def _grad_y(self,
                x: jnp.ndarray,
                y: jnp.ndarray,
                c: Union[float, jnp.ndarray],
                beta: Union[float, jnp.ndarray],
                bandwidth: Union[float, jnp.ndarray]) -> Union[float, jnp.ndarray]:
        diff = (x - y) / bandwidth
        return - diff / bandwidth * beta \
               * jnp.power(c + 0.5 * jnp.sum(jnp.square(diff), axis=-1), beta - 1)

    @partial(jit, static_argnums=(0,))
    def _diag_grad_xy(self,
                      x: jnp.ndarray,
                      y: jnp.ndarray,
                      c: Union[float, jnp.ndarray],
                      beta: Union[float, jnp.ndarray],
                      bandwidth: Union[float, jnp.ndarray]) -> Union[float, jnp.ndarray]:
        diff = (x - y) / bandwidth
        base = c + 0.5 * jnp.sum(jnp.square(diff), axis=-1)
        return (- beta * jnp.power(base, beta - 1) + beta * (beta - 1) * diff ** 2 * jnp.power(base, beta - 2)) \
               / bandwidth ** 2


def median_bandwidth_update(vals: jnp.ndarray) -> float:
    # Note jax.numpy.median scales much worse than numpy.median,
    # thus this method is not currently recommended
    dist_mat = l2_distance_matrix(vals)
    return jnp.median(dist_mat) / jnp.sqrt(2 * jnp.log(dist_mat.shape[0]))


def mean_bandwidth_update(vals: jnp.ndarray) -> float:
    dist_mat = l2_distance_matrix(vals)
    return jnp.mean(dist_mat) / jnp.sqrt(2 * jnp.log(dist_mat.shape[0]))




########################################################################################################################
# Module: tests/test_utils.py
# Description: Tests for mocat utils
#
# Web: https://github.com/SamDuffield/mocat-jax
########################################################################################################################

import unittest
import warnings

import jax.numpy as jnp
from jax.scipy.stats import multivariate_normal
from jax import vmap, random
import numpy.testing as npt

from mocat.src import utils
from mocat.src.core import cdict


class TestLeaveOneOut(unittest.TestCase):
    n = 10

    def test_0(self):
        inds0 = utils.leave_one_out_indices(self.n, 0)
        npt.assert_array_equal(inds0, jnp.arange(1, self.n + 1))

    def test_2(self):
        inds2 = utils.leave_one_out_indices(self.n, 2)
        npt.assert_array_equal(inds2[:2], jnp.arange(2))
        npt.assert_array_equal(inds2[2:], jnp.arange(3, self.n + 1))

    def test_n_plus_1(self):
        inds_n_plus_1 = utils.leave_one_out_indices(self.n, self.n + 1)
        npt.assert_array_equal(inds_n_plus_1, jnp.arange(self.n))


class TestGaussianPotential(unittest.TestCase):

    def test_n1_d1(self):
        x = jnp.array([7.])
        npt.assert_array_equal(utils.gaussian_potential(x), -multivariate_normal.logpdf(x, 0., 1.))

        m = jnp.array([1.])
        npt.assert_array_equal(utils.gaussian_potential(x, m), -multivariate_normal.logpdf(x, m, 1.))

        prec = jnp.array([[2.]])
        # test diag
        npt.assert_array_equal(utils.gaussian_potential(x, prec=prec[0]), -multivariate_normal.logpdf(x, 0, 1 / prec))
        npt.assert_array_almost_equal(utils.gaussian_potential(x, m, prec[0]),
                                      -multivariate_normal.logpdf(x, m, 1 / prec), decimal=4)
        # test full (omits norm constant)
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            npt.assert_array_equal(utils.gaussian_potential(x, prec=prec), 0.5 * x ** 2 * prec)
            npt.assert_array_equal(utils.gaussian_potential(x, m, prec), 0.5 * (x - m) ** 2 * prec)

    def test_n1_d5(self):
        x = jnp.ones(5)
        npt.assert_array_equal(utils.gaussian_potential(x), -multivariate_normal.logpdf(x, jnp.zeros_like(x), 1.))

        m = 3.
        npt.assert_array_equal(utils.gaussian_potential(x, m), -multivariate_normal.logpdf(x, m * jnp.ones_like(x), 1.))

        m = jnp.ones(5) * 3.
        npt.assert_array_equal(utils.gaussian_potential(x, m), -multivariate_normal.logpdf(x, m, 1.))

        # diagonal precision
        prec = jnp.eye(5) * 2
        # test diag
        npt.assert_array_almost_equal(utils.gaussian_potential(x, prec=jnp.diag(prec)),
                                      -multivariate_normal.logpdf(x, jnp.zeros_like(x), jnp.linalg.inv(prec)), decimal=5)
        npt.assert_array_almost_equal(utils.gaussian_potential(x, m, prec=jnp.diag(prec), det_prec=2 ** 5),
                                      -multivariate_normal.logpdf(x, m, jnp.linalg.inv(prec)), decimal=5)
        # test full (omits norm constant)
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            npt.assert_array_equal(utils.gaussian_potential(x, prec=prec), 0.5 * x.T @ prec @ x)
            npt.assert_array_equal(utils.gaussian_potential(x, m, prec), 0.5 * (x - m).T @ prec @ (x - m))
        # test full with det
        npt.assert_array_almost_equal(utils.gaussian_potential(x, prec=prec, det_prec=2 ** 5),
                                      -multivariate_normal.logpdf(x, jnp.zeros_like(x), jnp.linalg.inv(prec)), decimal=5)
        npt.assert_array_almost_equal(utils.gaussian_potential(x, m, prec=prec, det_prec=2 ** 5),
                                      -multivariate_normal.logpdf(x, m, jnp.linalg.inv(prec)), decimal=5)

        # non-diagonal precision
        sqrt_prec = jnp.arange(25).reshape(5, 5) / 100 + jnp.eye(5)
        prec = sqrt_prec @ sqrt_prec.T
        # test full (omits norm constant)
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            npt.assert_array_equal(utils.gaussian_potential(x, prec=prec), 0.5 * x.T @ prec @ x)
            npt.assert_array_equal(utils.gaussian_potential(x, m, prec), 0.5 * (x - m).T @ prec @ (x - m))
        # test full with det
        npt.assert_array_almost_equal(utils.gaussian_potential(x, prec=prec, det_prec=jnp.linalg.det(prec)),
                                      -multivariate_normal.logpdf(x, jnp.zeros_like(x), jnp.linalg.inv(prec)), decimal=5)
        npt.assert_array_almost_equal(utils.gaussian_potential(x, m, prec=prec, det_prec=jnp.linalg.det(prec)),
                                      -multivariate_normal.logpdf(x, m, jnp.linalg.inv(prec)), decimal=5)

    def test_n5_d2(self):
        x = jnp.ones((5, 2))
        npt.assert_array_equal(utils.gaussian_potential(x),
                               jnp.repeat(-multivariate_normal.logpdf(x[0], jnp.zeros(x.shape[-1]), 1.), 5))

        m = 3.
        npt.assert_array_equal(utils.gaussian_potential(x, m),
                               jnp.repeat(-multivariate_normal.logpdf(x[0], m * jnp.ones(x.shape[-1]), 1.), 5))

        m = jnp.ones(2) * 3.
        npt.assert_array_equal(utils.gaussian_potential(x, m),
                               jnp.repeat(-multivariate_normal.logpdf(x[0], m, 1.), 5))

        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            sqrt_prec = jnp.array([[5., 0.], [2., 3.]])
            npt.assert_array_equal(utils.gaussian_potential(x, sqrt_prec=sqrt_prec),
                                   jnp.repeat(0.5 * x[0].T @ sqrt_prec @ sqrt_prec.T @ x[0], 5))
            npt.assert_array_equal(utils.gaussian_potential(x, m, sqrt_prec=sqrt_prec),
                                   jnp.repeat(0.5 * (x[0] - m).T @ sqrt_prec @ sqrt_prec.T @ (x[0] - m), 5))


class TestLeapfrog(unittest.TestCase):
    stepsize = 0.1
    leapfrog_steps = 3

    start_state = cdict(value=jnp.zeros(2),
                        prior_potential=jnp.array(0.),
                        grad_prior_potential=jnp.array([0., 0.]),
                        likelihood_potential=jnp.array(0.),
                        grad_likelihood_potential=jnp.array([1., 2.]),
                        potential=jnp.array(0.),
                        grad_potential=jnp.array([1., 2.]),
                        momenta=jnp.ones(2),
                        auxiliary_float=5.,
                        auxiliary_0darray=jnp.array(0.),
                        auxiliary_2darray=jnp.zeros(2))

    def test_leapfrog(self):
        full_state = utils.leapfrog(lambda x, _: (0., jnp.zeros(2)),
                                    lambda x, _: (0., x),
                                    self.start_state,
                                    self.stepsize,
                                    random.split(random.PRNGKey(0), self.leapfrog_steps))
        out_state = full_state[-1]

        npt.assert_array_equal(out_state.value, jnp.array([0.28120947, 0.266409]))
        npt.assert_array_equal(out_state.grad_potential, jnp.array([0.28120947, 0.266409]))
        npt.assert_array_equal(out_state.momenta, jnp.array([0.9075344, 0.8597696]))

    def test_all_leapfrog(self):
        full_state = utils.leapfrog(lambda x, _: (0., jnp.zeros(2)),
                                    lambda x, _: (0., x),
                                    self.start_state,
                                    self.stepsize,
                                    random.split(random.PRNGKey(0), self.leapfrog_steps))

        npt.assert_array_equal(full_state.value.shape, (self.leapfrog_steps + 1, 2))
        npt.assert_array_equal(full_state.momenta.shape, (2 * self.leapfrog_steps + 1, 2))

        npt.assert_array_equal(full_state.value[0], jnp.zeros(2))
        npt.assert_array_equal(full_state.value[-1], jnp.array([0.28120947, 0.266409]))

        npt.assert_array_equal(full_state.momenta[0], jnp.ones(2))
        npt.assert_array_equal(full_state.momenta[-1], jnp.array([0.9075344, 0.8597696]))


class TestStackedWhile(unittest.TestCase):

    def test_trival(self):
        fin_iter = 10
        cond_fun = lambda x, _: x < fin_iter
        body_fun = lambda x, _: (x + 1, None)

        stack = utils.while_loop_stacked(cond_fun, body_fun, (0, None), 100)

        npt.assert_array_equal(stack, jnp.arange(1, fin_iter + 1))


class TestBisect(unittest.TestCase):

    def test_squared_increasing(self):
        bisect_fun = lambda x: x ** 2 - 10
        increasing_init_bounds = jnp.array([0, 1e2])

        increasing_bound, increasing_evals, increasing_iter = utils.bisect(bisect_fun, increasing_init_bounds)

        npt.assert_(jnp.any(jnp.abs(increasing_evals < 1e-3)))

    def test_squared_decreasing(self):
        bisect_fun = lambda x: x ** 2 - 10
        decreasing_init_bounds = jnp.array([-1e1, 0])

        decreasing_bound, decreasing_evals, decreasing_iter = utils.bisect(bisect_fun, decreasing_init_bounds)

        npt.assert_(jnp.any(jnp.abs(decreasing_evals < 1e-3)))


class TestBFGS(unittest.TestCase):
    hess_inv = jnp.array([[1., -0.5], [-0.5, 2.0]])
    # hess_inv = jnp.array([[1., 0.], [0., 5.0]])
    hess = jnp.linalg.inv(hess_inv)
    hess_inv_sqrt = jnp.linalg.cholesky(hess_inv)
    hess_sqrt = jnp.linalg.cholesky(hess)

    grad_gauss_pot = lambda h, x: h @ x

    vals = random.normal(random.PRNGKey(0), shape=(10, 2))
    grads = vmap(grad_gauss_pot, in_axes=(None, 0))(hess, vals)

    def test_not_pd(self):
        z = jnp.array([4.2, -3.7])
        init_hessian_sqrt_diag = jnp.array([0.3, 1.5])
        init_hessian_sqrt_diag = jnp.diag(self.hess_sqrt)
        # init_hessian_sqrt_diag = jnp.ones(2) * 0.1

        ps, qs, us, ts = utils.bfgs_sqrt_pqut(self.vals, self.grads, init_hessian_sqrt_diag)

        # Test transpose prod
        hess_inv_sqrt_t_z = utils.bfgs_sqrt_transpose_prod(ps, qs, z, 1/init_hessian_sqrt_diag)
        z2 = z.copy()
        for i in range(len(ps)-1 , -1, -1):
            z2 = (jnp.eye(2) - jnp.outer(qs[i], ps[i])) @ z2
        z2 = 1/init_hessian_sqrt_diag * z2
        npt.assert_array_almost_equal(hess_inv_sqrt_t_z, z2, decimal=3)

        # Test prod
        hess_inv_z = utils.bfgs_sqrt_prod(ps, qs, hess_inv_sqrt_t_z, 1/init_hessian_sqrt_diag)
        z3 = 1/init_hessian_sqrt_diag * z2.copy()
        for i in range(len(ps)):
            z3 = (jnp.eye(2) - jnp.outer(ps[i], qs[i])) @ z3
        npt.assert_array_almost_equal(hess_inv_z, z3, decimal=3)

        # Test accurate hessian inverse mvp
        npt.assert_array_almost_equal(self.hess_inv @ z, hess_inv_z, decimal=3)

        # Test accurate hessian mvp
        hess_sqrt_t_z = utils.bfgs_sqrt_transpose_prod(us, ts, z, init_hessian_sqrt_diag)
        hess_z = utils.bfgs_sqrt_prod(us, ts, hess_sqrt_t_z, init_hessian_sqrt_diag)
        npt.assert_array_almost_equal(self.hess @ z, hess_z, decimal=3)

        # Test determinant
        hess_inv_det = utils.bfgs_sqrt_det(ps, qs, 1/init_hessian_sqrt_diag) ** 2
        npt.assert_almost_equal(hess_inv_det, jnp.linalg.det(self.hess_inv), decimal=3)
        hess_det = utils.bfgs_sqrt_det(us, ts, init_hessian_sqrt_diag) ** 2
        npt.assert_almost_equal(hess_det, jnp.linalg.det(self.hess), decimal=3)


    def test_pd(self):
        z = jnp.array([4.2, -3.7])
        init_hessian_sqrt_diag = jnp.array([0.3, 1.5])
        init_hessian_sqrt_diag = jnp.diag(self.hess_sqrt)
        # init_hessian_sqrt_diag = jnp.ones(2) * 0.1

        ps, qs, us, ts = utils.bfgs_sqrt_pqut(self.vals, self.grads, init_hessian_sqrt_diag, r=0.5)

        # Test transpose prod
        hess_inv_sqrt_t_z = utils.bfgs_sqrt_transpose_prod(ps, qs, z, 1/init_hessian_sqrt_diag)
        z2 = z.copy()
        for i in range(len(ps)-1 , -1, -1):
            z2 = (jnp.eye(2) - jnp.outer(qs[i], ps[i])) @ z2
        z2 = 1/init_hessian_sqrt_diag * z2
        npt.assert_array_almost_equal(hess_inv_sqrt_t_z, z2, decimal=3)

        # Test prod
        hess_inv_z = utils.bfgs_sqrt_prod(ps, qs, hess_inv_sqrt_t_z, 1/init_hessian_sqrt_diag)
        z3 = 1/init_hessian_sqrt_diag * z2.copy()
        for i in range(len(ps)):
            z3 = (jnp.eye(2) - jnp.outer(ps[i], qs[i])) @ z3
        npt.assert_array_almost_equal(hess_inv_z, z3, decimal=3)

        # Test accurate hessian inverse mvp
        npt.assert_array_almost_equal(self.hess_inv @ z, hess_inv_z, decimal=3)

        # Test accurate hessian mvp
        hess_sqrt_t_z = utils.bfgs_sqrt_transpose_prod(us, ts, z, init_hessian_sqrt_diag)
        hess_z = utils.bfgs_sqrt_prod(us, ts, hess_sqrt_t_z, init_hessian_sqrt_diag)
        npt.assert_array_almost_equal(self.hess @ z, hess_z, decimal=3)

        # Test determinant
        hess_inv_det = utils.bfgs_sqrt_det(ps, qs, 1/init_hessian_sqrt_diag) ** 2
        npt.assert_almost_equal(hess_inv_det, jnp.linalg.det(self.hess_inv), decimal=3)
        hess_det = utils.bfgs_sqrt_det(us, ts, init_hessian_sqrt_diag) ** 2
        npt.assert_almost_equal(hess_det, jnp.linalg.det(self.hess), decimal=3)

        # Test inverse Hessian positive definite - somewhat null and void as the Gaussian potential is convex
        def inv_hess_pd(vec):
            hess_inv_sqrt_t_vec = utils.bfgs_sqrt_transpose_prod(ps, qs, vec, 1/init_hessian_sqrt_diag)
            hess_inv_vec = utils.bfgs_sqrt_prod(ps, qs, hess_inv_sqrt_t_vec, 1 / init_hessian_sqrt_diag)
            return jnp.dot(vec, hess_inv_vec)
        ips = vmap(inv_hess_pd)(self.vals)
        npt.assert_array_less(-ips, 0)


if __name__ == '__main__':
    unittest.main()



########################################################################################################################
# Module: tests/test_abc_linear_gaussian.py
# Description: Tests for ABC sampling (intractable likelihoods).
#
# Web: https://github.com/SamDuffield/mocat
########################################################################################################################

import unittest

import jax.numpy as jnp
from jax.random import PRNGKey
import numpy.testing as npt

from mocat.src.sample import run
from mocat.src.abc.scenarios.linear_gaussian import LinearGaussian
from mocat.src.abc.importance import VanillaABC
from mocat.src.abc.mcmc import RandomWalkABC, RMMetropolisDiagStepsize
from mocat.src.abc.smc import MetropolisedABCSMCSampler


class TestLinearGaussianABC(unittest.TestCase):
    prior_mean = jnp.zeros(3)
    prior_covariance = jnp.array([[1.0, -0.6, 0.5], [-0.6, 0.7, 0.1], [0.5, -0.6, 3.0]])
    data = jnp.array([1.4, -2.5])
    likelihood_matrix = jnp.array([[1.0, 0.5, 0.], [0.2, 0.3, 1.5]])
    likelihood_covariance = jnp.array([[0.02, -0.01], [-0.01, 0.05]])

    kalman_gain = prior_covariance @ likelihood_matrix.T \
                  @ jnp.linalg.inv(likelihood_matrix @ prior_covariance @ likelihood_matrix.T + likelihood_covariance)
    posterior_mean = prior_mean + kalman_gain @ (data - likelihood_matrix @ prior_mean)
    posterior_covariance = prior_covariance - kalman_gain @ likelihood_matrix @ prior_covariance

    n = int(1e5)

    def __init__(self, *args, **kwargs):

        super().__init__(*args, **kwargs)
        self.scenario = LinearGaussian(prior_mean=self.prior_mean,
                                       prior_covariance=self.prior_covariance,
                                       data=self.data,
                                       likelihood_matrix=self.likelihood_matrix,
                                       likelihood_covariance=self.likelihood_covariance)

    def _test_mean(self,
                   val: jnp.ndarray,
                   precision: float = 3.0):
        samp_mean = val.mean(axis=0)
        self.assertLess(jnp.abs(self.posterior_mean - samp_mean).sum(), precision)

    def _test_cov(self,
                   val: jnp.ndarray,
                   precision: float = 3.0):
        samp_cov = jnp.cov(val.T)
        self.assertLess(jnp.abs(self.posterior_covariance - samp_cov).sum(), precision)


class TestVanillaLinearGaussianABC(TestLinearGaussianABC):

    def test_pre_threshold(self):
        threshold = 5.0
        sample = run(self.scenario, VanillaABC(threshold=threshold), n=self.n,
                     random_key=PRNGKey(0))
        self._test_mean(sample.value[sample.log_weight > -jnp.inf], 10.)
        self._test_cov(sample.value[sample.log_weight > -jnp.inf], 10.)

    def test_post_threshold(self):
        acceptance_rate = 0.1
        sampler = VanillaABC(acceptance_rate=acceptance_rate)
        sample = run(self.scenario, sampler, n=self.n,
                     random_key=PRNGKey(0))
        self._test_mean(sample.value[sample.log_weight > -jnp.inf], 10.)
        self._test_cov(sample.value[sample.log_weight > -jnp.inf], 10.)
        npt.assert_almost_equal(jnp.mean(sample.log_weight == 0), acceptance_rate, decimal=3)
        self.assertNotEqual(sampler.parameters.threshold, jnp.inf)


class TestMCMCLinearGaussianABC(TestLinearGaussianABC):
    threshold = 1.0

    def test_fixed_stepsize(self):
        stepsize = 0.1
        sample = run(self.scenario, RandomWalkABC(stepsize=stepsize, threshold=self.threshold), n=self.n,
                     random_key=PRNGKey(0))
        self._test_mean(sample.value)
        self._test_cov(sample.value)

    def test_adaptive_diag_stepsize(self):
        sampler = RandomWalkABC()
        sampler.tuning.target = 0.1
        sample = run(self.scenario, sampler, n=self.n,
                     random_key=PRNGKey(0), correction=RMMetropolisDiagStepsize())
        self._test_mean(sample.value)
        self._test_cov(sample.value)
        npt.assert_almost_equal(sample.alpha.mean(), sampler.tuning.target, decimal=1)
        npt.assert_array_almost_equal(sample.stepsize[-1],
                                      jnp.diag(jnp.cov(sample.value, rowvar=False)) / self.scenario.dim * 2.38 ** 2,
                                      decimal=1)


class TestMetSMCLinearGaussianABC(TestLinearGaussianABC):
    n = int(1e3)

    def test_threshold_preschedule(self):
        threshold_schedule = jnp.linspace(10, 0.1, 10)
        sampler = MetropolisedABCSMCSampler(threshold_schedule=threshold_schedule)
        sample = run(self.scenario, sampler, n=self.n,
                     random_key=PRNGKey(0))
        self._test_mean(sample.value[-1])
        self._test_cov(sample.value[-1])

    def test_adaptive(self):
        retain_parameter = 0.8
        sampler = MetropolisedABCSMCSampler(ess_threshold_retain=retain_parameter)
        sample = run(self.scenario, sampler,
                     n=self.n,
                     random_key=PRNGKey(0))
        self._test_mean(sample.value[-1])
        self._test_cov(sample.value[-1])



########################################################################################################################
# Module: transport/smc.py
# Description: SMC samplers for tempered sequence of distributions.
#
# Web: https://github.com/SamDuffield/mocat
########################################################################################################################

from typing import Tuple, Union, Type
from inspect import isclass

from jax import numpy as jnp, random, vmap
from jax.lax import scan, cond
from jax.scipy.special import logsumexp

from mocat.src.core import Scenario, cdict
from mocat.src.transport.sampler import TransportSampler
from mocat.src.mcmc.sampler import MCMCSampler, Correction, check_correction
from mocat.src.mcmc.metropolis import Metropolis
from mocat.src.utils import bisect
from mocat.src.metrics import log_ess_log_weight


class SMCSampler(TransportSampler):
    name = 'SMC Sampler'

    def startup(self,
                scenario: Scenario,
                n: int,
                initial_state: cdict,
                initial_extra: cdict,
                **kwargs) -> Tuple[cdict, cdict]:
        initial_state, initial_extra = super().startup(scenario, n, initial_state, initial_extra, **kwargs)
        if not hasattr(initial_extra, 'resample_bool'):
            initial_extra.resample_bool = True
        if not hasattr(initial_state, 'log_weight'):
            initial_state.log_weight = jnp.zeros(n)
        if not hasattr(initial_state, 'ess'):
            initial_state.ess = jnp.zeros(n) + n
        return initial_state, initial_extra

    def forward_proposal(self,
                         scenario: Scenario,
                         previous_state: cdict,
                         previous_extra: cdict,
                         random_key: jnp.ndarray) -> cdict:
        raise AttributeError(f'{self.name} forward_proposal not initiated')

    def adapt(self,
              previous_ensemble_state: cdict,
              previous_extra: cdict,
              new_ensemble_state: cdict,
              new_extra: cdict) -> Tuple[cdict, cdict]:
        new_ensemble_state.log_weight = previous_ensemble_state.log_weight + new_ensemble_state.log_weight
        return new_ensemble_state, new_extra

    def resample_criterion(self,
                           ensemble_state: cdict,
                           extra: cdict) -> bool:
        return True

    def resample(self,
                 ensemble_state: cdict,
                 random_key: jnp.ndarray) -> cdict:
        n = ensemble_state.value.shape[0]
        resampled_indices = random.categorical(random_key,
                                               ensemble_state.log_weight,
                                               shape=(n,))
        resampled_ensemble_state = ensemble_state[resampled_indices]
        resampled_ensemble_state.log_weight = jnp.zeros(n)
        resampled_ensemble_state.ess = jnp.zeros(n) + n
        return resampled_ensemble_state

    def update(self,
               scenario: Scenario,
               ensemble_state: cdict,
               extra: cdict) -> Tuple[cdict, cdict]:
        extra.iter = extra.iter + 1
        n = ensemble_state.value.shape[0]

        extra.resample_bool = self.resample_criterion(ensemble_state, extra)

        random_keys_all = random.split(extra.random_key, n + 2)
        extra.random_key = random_keys_all[-1]

        resampled_ensemble_state \
            = cond(extra.resample_bool,
                   lambda state: self.resample(state, random_keys_all[-2]),
                   lambda state: state,
                   ensemble_state)

        advanced_state = vmap(self.forward_proposal,
                              in_axes=(None, 0, None, 0))(scenario,
                                                          resampled_ensemble_state,
                                                          extra,
                                                          random_keys_all[:n])
        advanced_state, advanced_extra = self.adapt(resampled_ensemble_state, extra,
                                                    advanced_state, extra)

        return advanced_state, advanced_extra


class TemperedSMCSampler(SMCSampler):
    name = 'Tempered SMC Sampler'

    def __init__(self,
                 temperature_schedule: Union[None, jnp.ndarray] = None,
                 max_temperature: float = 1.,
                 max_iter: int = int(1e4),
                 **kwargs):
        self.max_iter = max_iter
        self.max_temperature = max_temperature
        self.temperature_schedule = temperature_schedule
        super().__init__(**kwargs)

    def __setattr__(self, key, value):
        if key == 'temperature_schedule':
            if value is None:
                self.next_temperature = self.next_temperature_adaptive
                if hasattr(self, 'temperature_schedule') and self.temperature_schedule is not None \
                        and self.max_iter == len(self.temperature_schedule):
                    self.max_iter = int(1e4)
            else:
                self.next_temperature = lambda state, extra: self.temperature_schedule[extra.iter]
                self.max_temperature = value[-1]
                self.max_iter = len(value)
        super().__setattr__(key, value)

    def startup(self,
                scenario: Scenario,
                n: int,
                initial_state: cdict,
                initial_extra: cdict,
                initiate_potential: bool = True,
                **kwargs) -> Tuple[cdict, cdict]:
        if not hasattr(scenario, 'prior_sample'):
            raise TypeError(f'Likelihood tempering requires scenario {scenario.name} to have prior_sample implemented')

        initial_state, initial_extra = super().startup(scenario, n, initial_state, initial_extra, **kwargs)

        random_keys = random.split(initial_extra.random_key, 2 * n + 1)

        initial_extra.random_key = random_keys[-1]
        if initiate_potential:
            if hasattr(scenario, 'prior_potential_and_grad'):
                initial_state.prior_potential, initial_state.grad_prior_potential \
                    = vmap(scenario.prior_potential_and_grad)(initial_state.value, random_keys[:n])
                initial_state.likelihood_potential, initial_state.grad_likelihood_potential \
                    = vmap(scenario.likelihood_potential_and_grad)(initial_state.value, random_keys[n:(2 * n)])
                initial_state.potential = initial_state.prior_potential
                initial_state.grad_potential = initial_state.grad_prior_potential
            else:
                initial_state.prior_potential = vmap(scenario.prior_potential)(initial_state.value, random_keys[:n])
                initial_state.likelihood_potential = vmap(scenario.likelihood_potential)(initial_state.value,
                                                                                         random_keys[n:(2 * n)])
                initial_state.potential = initial_state.prior_potential

        initial_state.temperature = jnp.zeros(n)
        initial_state.log_weight = jnp.zeros(n)
        initial_state.ess = jnp.zeros(n) + n
        initial_state.log_norm_constant = logsumexp(initial_state.log_weight, b=1 / n) * jnp.ones(n)

        scenario.temperature = 0.

        return initial_state, initial_extra

    def next_temperature_adaptive(self,
                                  ensemble_state: cdict,
                                  extra: cdict) -> float:
        raise AttributeError(f'{self.name} next_temperature_adaptive not initiated')

    def termination_criterion(self,
                              ensemble_state: cdict,
                              extra: cdict) -> bool:
        return jnp.logical_or(jnp.logical_or(ensemble_state.temperature[0] >= self.max_temperature,
                                             extra.iter >= self.max_iter), jnp.isnan(ensemble_state.value).mean() > 0.1)

    def clean_chain(self,
                    scenario: Scenario,
                    chain_ensemble_state: cdict) -> cdict:
        chain_ensemble_state.temperature = chain_ensemble_state.temperature[:, 0]
        scenario.temperature = float(chain_ensemble_state.temperature[-1])
        chain_ensemble_state.ess = chain_ensemble_state.ess[:, 0]
        chain_ensemble_state.log_norm_constant = chain_ensemble_state.log_norm_constant[:, 0]
        return chain_ensemble_state

    def log_weight(self,
                   previous_ensemble_state: cdict,
                   previous_extra: cdict,
                   new_ensemble_state: cdict,
                   new_extra: cdict) -> Union[float, jnp.ndarray]:
        return 0.

    def adapt(self,
              previous_ensemble_state: cdict,
              previous_extra: cdict,
              new_ensemble_state: cdict,
              new_extra: cdict) -> Tuple[cdict, cdict]:
        n = new_ensemble_state.value.shape[0]
        next_temperature = self.next_temperature(new_ensemble_state, new_extra)
        new_ensemble_state.temperature = jnp.ones(n) * next_temperature
        new_ensemble_state.log_weight = previous_ensemble_state.log_weight \
                                        + self.log_weight(previous_ensemble_state, previous_extra,
                                                          new_ensemble_state, new_extra)
        new_ensemble_state.ess = jnp.ones(n) * jnp.exp(log_ess_log_weight(new_ensemble_state.log_weight))
        new_ensemble_state.potential = new_ensemble_state.prior_potential \
                                       + next_temperature * new_ensemble_state.likelihood_potential
        if hasattr(new_ensemble_state, 'grad_potential'):
            new_ensemble_state.grad_potential \
                = new_ensemble_state.grad_prior_potential \
                  + next_temperature * new_ensemble_state.grad_likelihood_potential

        new_ensemble_state.log_norm_constant \
            = previous_ensemble_state.log_norm_constant \
              + logsumexp(new_ensemble_state.log_weight) \
              - logsumexp(previous_ensemble_state.log_weight)

        return new_ensemble_state, new_extra

    def update(self,
               scenario: Scenario,
               ensemble_state: cdict,
               extra: cdict) -> Tuple[cdict, cdict]:
        scenario.temperature = ensemble_state.temperature[0]
        advanced_state, advanced_extra = super().update(scenario, ensemble_state, extra)
        return advanced_state, advanced_extra


class MetropolisedSMCSampler(TemperedSMCSampler):
    name = "Metropolised SMC Sampler"

    def __init__(self,
                 mcmc_sampler: Union[MCMCSampler, Type[MCMCSampler]],
                 mcmc_correction: Union[Correction, Type[Correction], str] = 'sampler_default',
                 mcmc_steps: int = 1,
                 max_iter: int = int(1e4),
                 temperature_schedule: Union[None, jnp.ndarray] = None,
                 max_temperature: float = 1.,
                 ess_threshold_retain: float = 0.9,
                 ess_threshold_resample: float = 0.5,
                 bisection_tol: float = 1e-5,
                 max_bisection_iter: int = 1000,
                 **kwargs):
        if temperature_schedule is not None:
            if temperature_schedule[0] == 0.:
                temperature_schedule = temperature_schedule[1:]

        super().__init__(max_iter=max_iter, temperature_schedule=temperature_schedule, max_temperature=max_temperature,
                         **kwargs)
        if isclass(mcmc_sampler):
            mcmc_sampler = mcmc_sampler()
        self.mcmc_sampler = mcmc_sampler
        if mcmc_correction != 'sampler_default':
            self.mcmc_sampler.correction = mcmc_correction
        self.parameters.mcmc_steps = mcmc_steps

        self.parameters.ess_threshold_retain = ess_threshold_retain
        self.parameters.ess_threshold_resample = ess_threshold_resample
        self.parameters.bisection_tol = bisection_tol
        self.parameters.max_bisection_iter = max_bisection_iter

    def __setattr__(self, key, value):
        if key == 'temperature_schedule':
            if value is not None and value[0] == 0.:
                value = value[1:]
        super().__setattr__(key, value)

    def startup(self,
                scenario: Scenario,
                n: int,
                initial_state: cdict,
                initial_extra: cdict,
                **kwargs) -> Tuple[cdict, cdict]:

        self.mcmc_sampler.correction = check_correction(self.mcmc_sampler.correction)

        initial_state, initial_extra = super().startup(scenario, n, initial_state, initial_extra, **kwargs)

        first_temp = self.next_temperature(initial_state, initial_extra)
        scenario.temperature = first_temp
        initial_state.temperature += first_temp
        initial_state.potential = initial_state.prior_potential + first_temp * initial_state.likelihood_potential
        if hasattr(initial_state, 'grad_potential'):
            initial_state.grad_potential = initial_state.grad_prior_potential \
                                           + first_temp * initial_state.grad_likelihood_potential

        initial_state.log_weight = - first_temp * initial_state.likelihood_potential
        initial_state.ess = jnp.repeat(jnp.exp(log_ess_log_weight(initial_state.log_weight)), n)
        initial_state.log_norm_constant = logsumexp(initial_state.log_weight, b=1 / n) * jnp.ones(n)

        initial_state, initial_extra = vmap(
            lambda state: self.mcmc_sampler.startup(scenario,
                                                    n,
                                                    state,
                                                    initial_extra))(initial_state)
        initial_extra = initial_extra[0]
        return initial_state, initial_extra

    def resample_criterion(self,
                           ensemble_state: cdict,
                           extra: cdict) -> bool:
        return ensemble_state.ess[0] <= (extra.parameters.ess_threshold_resample * len(ensemble_state.value))

    @staticmethod
    def log_ess(previous_log_weight: jnp.ndarray,
                current_temperature: float,
                new_temperature: float,
                likelihood_potential: jnp.ndarray) -> float:
        log_weight = previous_log_weight - (new_temperature - current_temperature) * likelihood_potential
        return log_ess_log_weight(log_weight)

    def next_temperature_adaptive(self,
                                  ensemble_state: cdict,
                                  extra: cdict) -> float:
        temperature_bounds = jnp.array([ensemble_state.temperature[0], self.max_temperature])
        likelihood_potential = ensemble_state.likelihood_potential
        log_n_samp_threshold = jnp.log(ensemble_state.ess[0] * self.parameters.ess_threshold_retain)

        bisect_out_bounds, bisect_out_evals, bisect_out_iter \
            = bisect(lambda x: self.log_ess(ensemble_state.log_weight,
                                            temperature_bounds[0],
                                            x,
                                            likelihood_potential) - log_n_samp_threshold,
                     temperature_bounds,
                     max_iter=self.parameters.max_bisection_iter,
                     tol=self.parameters.bisection_tol)
        return bisect_out_bounds[jnp.argmin(jnp.abs(bisect_out_evals))]

    def clean_mcmc_chain(self,
                         chain_state: cdict,
                         chain_extra: cdict) -> Tuple[cdict, cdict]:
        clean_state = chain_state[-1]
        clean_extra = chain_extra[-1]
        clean_state.alpha = chain_state.alpha.mean()
        clean_extra.parameters = chain_extra.parameters[-1]
        return clean_state, clean_extra

    def forward_proposal(self,
                         scenario: Scenario,
                         state: cdict,
                         extra: cdict,
                         random_key: jnp.ndarray) -> cdict:

        def mcmc_kernel(previous_carry: Tuple[cdict, cdict],
                        _: None) -> Tuple[Tuple[cdict, cdict], Tuple[cdict, cdict]]:
            new_carry = self.mcmc_sampler.update(scenario, *previous_carry)
            return new_carry, new_carry

        extra.random_key = random_key

        start_state, start_extra = self.mcmc_sampler.startup(scenario,
                                                             extra.parameters.mcmc_steps,
                                                             state,
                                                             extra)

        final_carry, chain = scan(mcmc_kernel,
                                  (start_state, start_extra),
                                  None,
                                  length=self.parameters.mcmc_steps)

        advanced_state, advanced_extra = self.clean_mcmc_chain(chain[0], chain[1])

        advanced_state.prior_potential = scenario.prior_potential(advanced_state.value, advanced_extra.random_key)
        advanced_state.likelihood_potential = (advanced_state.potential - advanced_state.prior_potential) \
                                              / scenario.temperature
        return advanced_state

    def log_weight(self,
                   previous_ensemble_state: cdict,
                   previous_extra: cdict,
                   new_ensemble_state: cdict,
                   new_extra: cdict) -> jnp.ndarray:
        return - (new_ensemble_state.temperature[0] - previous_ensemble_state.temperature[0]) \
               * new_ensemble_state.likelihood_potential


class RMMetropolisedSMCSampler(MetropolisedSMCSampler):
    def __init__(self,
                 mcmc_sampler: Union[MCMCSampler, Type[MCMCSampler]],
                 mcmc_correction: Union[Correction, Type[Correction], str] = Metropolis,
                 mcmc_steps: int = 1,
                 max_iter: int = int(1e4),
                 temperature_schedule: Union[None, jnp.ndarray] = None,
                 max_temperature: float = 1.,
                 ess_threshold_retain: float = 0.9,
                 ess_threshold_resample: float = 0.5,
                 bisection_tol: float = 1e-5,
                 max_bisection_iter: int = 1000,
                 rm_stepsize: float = 1.,
                 **kwargs):
        super().__init__(mcmc_sampler=mcmc_sampler, mcmc_correction=mcmc_correction, mcmc_steps=mcmc_steps,
                         max_iter=max_iter, temperature_schedule=temperature_schedule, max_temperature=max_temperature,
                         ess_threshold_retain=ess_threshold_retain, ess_threshold_resample=ess_threshold_resample,
                         bisection_tol=bisection_tol, max_bisection_iter=max_bisection_iter, **kwargs)
        self.parameters.rm_stepsize = rm_stepsize

    def startup(self,
                scenario: Scenario,
                n: int,
                initial_state: cdict,
                initial_extra: cdict,
                **kwargs) -> Tuple[cdict, cdict]:
        initial_state, initial_extra = super().startup(scenario, n, initial_state, initial_extra, **kwargs)
        initial_state.stepsize = jnp.ones(n) * initial_extra.parameters.stepsize
        return initial_state, initial_extra

    def adapt(self,
              previous_ensemble_state: cdict,
              previous_extra: cdict,
              new_ensemble_state: cdict,
              new_extra: cdict) -> Tuple[cdict, cdict]:
        new_ensemble_state, new_extra = super().adapt(previous_ensemble_state, previous_extra,
                                                      new_ensemble_state, new_extra)
        log_stepsize = jnp.log(new_extra.parameters.stepsize)
        alpha_mean = jnp.average(new_ensemble_state.alpha,
                                 weights=jnp.exp(new_ensemble_state.log_weight - new_ensemble_state.log_weight.max()))
        new_log_stepsize = log_stepsize + new_extra.parameters.rm_stepsize \
                           * (alpha_mean - self.mcmc_sampler.tuning.target)
        new_extra.parameters.stepsize = jnp.exp(new_log_stepsize)

        new_ensemble_state.stepsize = jnp.ones(new_ensemble_state.value.shape[0]) * new_extra.parameters.stepsize
        return new_ensemble_state, new_extra

    def clean_chain(self,
                    scenario: Scenario,
                    chain_ensemble_state: cdict) -> cdict:
        chain_ensemble_state = super().clean_chain(scenario, chain_ensemble_state)
        chain_ensemble_state.stepsize = chain_ensemble_state.stepsize[:, 0]
        return chain_ensemble_state





from mocat.src.utils import leapfrog
from mocat.src.utils import gaussian_potential
from mocat.src.utils import leave_one_out_indices
from mocat.src.utils import _while_loop_stacked
from mocat.src.utils import while_loop_stacked
from mocat.src.utils import bisect
from mocat.src.utils import extract_dimension
from mocat.src.utils import reset_covariance
from mocat.src.utils import bfgs_sqrt_pqut
from mocat.src.utils import bfgs_sqrt_prod
from mocat.src.utils import bfgs_sqrt_transpose_prod
from mocat.src.utils import bfgs_sqrt_det
from mocat.src.utils import l2_distance_matrix
from mocat.src.utils import gc_matrix

from mocat.src.mcmc.qn_underdamped import bfgs_leapfrog

from mocat.src.core import impl_checkable
from mocat.src.core import is_implemented




########################################################################################################################
# Module: tests/test_abc_linear_gaussian.py
# Description: Tests for ABC sampling (intractable likelihoods).
#
# Web: https://github.com/SamDuffield/mocat
########################################################################################################################

import unittest

import jax.numpy as jnp
from jax.random import PRNGKey
import numpy.testing as npt

from mocat.src.sample import run
from mocat.src.abc.scenarios.gk import GKTransformedUniformPrior, GKOnlyATransformedUniformPrior
from mocat.src.abc.importance import VanillaABC
from mocat.src.abc.mcmc import RandomWalkABC, RMMetropolisDiagStepsize
from mocat.src.abc.smc import MetropolisedABCSMCSampler


class GKThinOrder(GKTransformedUniformPrior):
    num_thin: int = 100
    n_unsummarised_data: int = 1000

    def summarise_data(self,
                       data: jnp.ndarray):
        order_stats = data.sort()
        thin_inds = jnp.linspace(0, len(data), self.num_thin, endpoint=False, dtype='int32')
        return order_stats[thin_inds]


class GKOnlyAThinOrder(GKOnlyATransformedUniformPrior):
    num_thin: int = 100
    n_unsummarised_data: int = 1000

    def summarise_data(self,
                       data: jnp.ndarray):
        order_stats = data.sort()
        thin_inds = jnp.linspace(0, len(data), self.num_thin, endpoint=False, dtype='int32')
        return order_stats[thin_inds]


class TestGK(unittest.TestCase):

    n = int(1e5)
    onlyA = False

    def __init__(self, *args, **kwargs):

        super().__init__(*args, **kwargs)
        if self.onlyA:
            self.scenario = GKOnlyAThinOrder()
            self.true_constrained_params = jnp.array([3.])
        else:
            self.scenario = GKThinOrder()
            self.true_constrained_params = jnp.array([3., 1., 2., 0.5])
        self.true_unconstrained_params = self.scenario.unconstrain(self.true_constrained_params)

        self.scenario.data = self.scenario.likelihood_sample(self.true_unconstrained_params,
                                                             random_key=PRNGKey(0))

    def _test_mean(self,
                   vals: jnp.ndarray,
                   precision: float = 3.0):
        samp_mean = vals.mean(axis=0)
        self.assertLess(jnp.abs(self.true_unconstrained_params - samp_mean).sum(), precision)


class TestVanillaGK(TestGK):

    def test_pre_threshold(self):
        threshold = 5.0
        sample = run(self.scenario, VanillaABC(threshold=threshold), n=self.n,
                     random_key=PRNGKey(0))
        self._test_mean(sample.value[sample.log_weight > -jnp.inf], 10.)

    def test_post_threshold(self):
        acceptance_rate = 0.1
        sampler = VanillaABC(acceptance_rate=acceptance_rate)
        sample = run(self.scenario, sampler, n=self.n,
                     random_key=PRNGKey(0))
        self._test_mean(sample.value[sample.log_weight > -jnp.inf], 10.)
        npt.assert_almost_equal(jnp.mean(sample.log_weight == 0), acceptance_rate, decimal=3)
        self.assertNotEqual(sampler.parameters.threshold, jnp.inf)


class TestVanillaGKOnlyA(TestGK):
    onlyA = True

    def test_pre_threshold(self):
        threshold = 1000.
        sample = run(self.scenario, VanillaABC(threshold=threshold), n=self.n,
                     random_key=PRNGKey(0))
        self._test_mean(sample.value[sample.log_weight > -jnp.inf], 10.)

    def test_post_threshold(self):
        acceptance_rate = 0.1
        sampler = VanillaABC(acceptance_rate=acceptance_rate)
        sample = run(self.scenario, sampler, n=self.n,
                     random_key=PRNGKey(0))
        self._test_mean(sample.value[sample.log_weight > -jnp.inf], 10.)
        npt.assert_almost_equal(jnp.mean(sample.log_weight == 0), acceptance_rate, decimal=3)
        self.assertNotEqual(sampler.parameters.threshold, jnp.inf)


class TestMCMCGK(TestGK):

    def test_adaptive_diag_stepsize(self):
        sampler = RandomWalkABC()
        sampler.tuning.target = 0.1
        sample = run(self.scenario, sampler, n=self.n,
                     random_key=PRNGKey(0), correction=RMMetropolisDiagStepsize(rm_stepsize_scale=0.01))
        self._test_mean(sample.value)
        npt.assert_almost_equal(sample.alpha.mean(), sampler.tuning.target, decimal=1)


class TestMCMCGKOnlyA(TestGK):

    onlyA = True

    def test_adaptive_diag_stepsize(self):
        sampler = RandomWalkABC()
        sampler.tuning.target = 0.1
        sample = run(self.scenario, sampler, n=self.n,
                     random_key=PRNGKey(0), correction=RMMetropolisDiagStepsize(rm_stepsize_scale=0.1))
        self._test_mean(sample.value)
        npt.assert_almost_equal(sample.alpha.mean(), sampler.tuning.target, decimal=1)


class TestMetSMCGK(TestGK):
    n = int(1e3)

    def test_threshold_preschedule(self):
        threshold_schedule = jnp.linspace(100., 10., 10)
        sampler = MetropolisedABCSMCSampler(threshold_schedule=threshold_schedule)
        sample = run(self.scenario, sampler, n=self.n,
                     random_key=PRNGKey(0))
        self._test_mean(sample.value[-1])

    def test_adaptive(self):
        retain_parameter = 0.8
        sampler = MetropolisedABCSMCSampler(ess_threshold_retain=retain_parameter, termination_alpha=0.1)
        sample = run(self.scenario, sampler, n=self.n,
                     random_key=PRNGKey(0))
        self._test_mean(sample.value[-1])


class TestMetSMCGKOnlyA(TestGK):
    n = int(1e3)
    onlyA = True

    def test_threshold_preschedule(self):
        threshold_schedule = jnp.linspace(100., 10., 10)
        sampler = MetropolisedABCSMCSampler(threshold_schedule=threshold_schedule)
        sample = run(self.scenario, sampler, n=self.n,
                     random_key=PRNGKey(0))
        self._test_mean(sample.value[-1])

    def test_adaptive(self):
        retain_parameter = 0.8
        sampler = MetropolisedABCSMCSampler(ess_threshold_retain=retain_parameter, termination_alpha=0.1)
        sample = run(self.scenario, sampler, n=self.n,
                     random_key=PRNGKey(0))
        self._test_mean(sample.value[-1])
